/***********************************************************************/
/*                                                                     */
/*   svm_struct_api.c       (modified for PCFG parsing)                */
/*                                                                     */
/*   Definition of API for attaching implementing SVM learning of      */
/*   structures (e.g. parsing, multi-label classification, HMM)        */ 
/*                                                                     */
/*   Based on CKY parser by Mark Johnson                               */
/*                                                                     */
/*   Author: Thorsten Joachims                                         */
/*   Date: 12.07.04                                                    */
/*                                                                     */
/*   Copyright (c) 2004  Thorsten Joachims - All rights reserved       */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include <string.h>
#include "svm_struct/svm_struct_common.h"
#include "svm_struct_api.h"
#include "svm_struct/svm_struct_learn.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "local-trees.h"	/* local tree count format */
#include "mmm.h"		/* memory debugger */
#include "hash-string.h" 	/* hash tables and string-index tables */
#include "hash-templates.h"
#include "vindex.h"
#include "tree.h"
#include "ledge.h"
#include "grammar.h"
#include "hash.h"
#include "hash-templates.h"

#define RAND_SEED	time(0)

#define CKY_MAX         0
#define CKY_STRATIFIED  1

#define NO_ADD_NEW_SYMBOLS 0
#define ADD_NEW_SYMBOLS 1

#define CKY_INITSTRAT   200  /* initial size of hash table for strat chells */

#define MAXUNARYDEPTH   2    /* maximum number of unary rules to apply
				in chain */

#define CHART_SIZE(n)			(n)*((n)+1)/2
#define CHART_ENTRY(chart, i, j)	chart[(j)*((j)-1)/2+(i)]

typedef struct strat_cell {
  unsigned int   correct;
  unsigned int   size;
  struct bintree tree;
  FLOAT		 prob;
  struct strat_cell     *next;
} *strat_cell;

HASH_HEADER(strat_hash, long, strat_cell)
HASH_CODE(strat_hash, long, strat_cell, IDENTITY, NEQ, IDENTITY, NO_OP, NULL, NO_OP)

typedef struct chart_cell {
  struct bintree tree;
  FLOAT		 prob;
  unsigned int   nalt;
  int            present2;
  struct bintree tree2;
  FLOAT		 prob2;
  unsigned int   nalt2;
  strat_cell     strat_list;
  strat_hash     strat_hash;
} *chart_cell;


chart_cell
make_chart_cell(si_index label, bintree left, bintree right,
		FLOAT prob, unsigned int nalt, int correct, int mode)
{
  chart_cell c = MALLOC(sizeof(struct chart_cell));
  c->tree.label = label;
  c->tree.left = left;
  c->tree.right = right;
  c->prob = prob;
  c->nalt = nalt;
  c->present2 = 0;
  c->tree2.label = label;
  c->tree2.left = NULL;
  c->tree2.right = NULL;
  c->prob2 = 0;
  c->nalt2 = nalt;

  c->strat_list = NULL;
  c->strat_hash = NULL;
  if(mode == CKY_STRATIFIED) {
    c->strat_hash = make_strat_hash(CKY_INITSTRAT);
  }
  return c;
}


/* chart_cell_free() frees the memory associated with this chart cell.
 * A chart cell has a tree associated with it, but since every tree
 * node is associated with exactly one chart cell, only free the
 * top-most node of each tree.
 */
void chart_cell_free(chart_cell c)
{
  strat_cell s,snext;
  /* printf("cprt=%ld %ld %ld %ld %ld %ld\n",c,c->strat_index,c->strat_correct,c->strat_size,c->strat_tree,c->strat_prob);fflush(stdout); */
  if(c->strat_hash) free_strat_hash(c->strat_hash);
  s=c->strat_list;
  while(s) {
    snext=s->next;
    FREE(s);
    s=snext;
  }
  FREE(c);
}

HASH_HEADER(sihashcc, si_index, chart_cell)
HASH_CODE(sihashcc, si_index, chart_cell, IDENTITY, NEQ, IDENTITY, NO_OP, 
	  NULL, chart_cell_free)

typedef sihashcc *chart;

void count_local_trees(const tree tree, vihashl localtree_ht);
	/* adds local tree counts from local trees in tree to localtree_ht */

void write_local_trees(FILE *fh, const vihashl localtree_ht, si_t si);
	/* writes local tree hash table to stdout */

int tree_eq(const tree t1, const tree t2);

grammar create_grammar(vihashl localtree_ht, vihashl weightid, si_t si);

void chart_free(chart c, size_t n);

SVECTOR *collect_phi(bintree parse, STRUCTMODEL *sm, 
		     size_t lpos, size_t *rpos, size_t start, size_t end);
SVECTOR *phi_urule(long weightid, bintree child, STRUCTMODEL *sm, 
		   int lpos, int rpos, int start, int end);
SVECTOR *phi_brule(long weightid, bintree left, bintree right, STRUCTMODEL *sm,
		   int lpos, int mpos, int rpos, int start, int end);
int add_feature(WORD *feat, int pos, long fnum, long weight);
int encode_number(WORD *feat,int pos,long basefnum,long number,double weight,
		  long a, long b, long c, long d, long e, long f);
double urule_value(urule rule, bintree child, STRUCTMODEL *sm, 
		   int lpos, int rpos, int start, int end);
double brule_value(brule rule, bintree left, bintree right, STRUCTMODEL *sm, 
		   int lpos, int mpos, int rpos, int start, int end);

chart cky(struct vindex terms, struct ledges *l, STRUCTMODEL *sm, si_t si, 
	  int mode);
chart cky_wrap(struct vindex terms, tree correct_tree, STRUCTMODEL *sm, 
	       si_t si);
chart cky_maxloss(struct vindex terms, PATTERN x, LABEL y, STRUCTMODEL *sm, 
		  STRUCT_LEARN_PARM *sparm, double *loss);

double fone(int correct, int trueledgenum, int predledgenum);


SAMPLE      read_struct_examples(char *file, STRUCT_LEARN_PARM *sparm)
{
  /* Reads struct examples and returns them in sample. The number of
     examples must be written into sample.n */
  SAMPLE   sample;  /* sample */
  EXAMPLE  *examples;
  long     totsen;
  long     n;       /* number of examples */

  tree     t;
  si_t     si;
  FILE     *fp;

  if(!sparm->si)
    si = make_si(100);
  else
    si = sparm->si;

  n=0;
  totsen=0;
  examples=(EXAMPLE *)MALLOC(sizeof(EXAMPLE)*100000);   /* hack */
  if ((fp = fopen (file, "r")) == NULL)
    { perror (file); exit (1); } 
  while ((t = readtree_root(fp, si))) {
    tree p = collapse_identical_unary(t); free_tree(t);
    if(sparm->parent_annotation) {
      t = annotate_with_parent(p, si); free_tree(p);
      p=t;
    }
    examples[n].y.parse = p;
    examples[n].y.si = si;
    examples[n].x.sentence = tree_terms(examples[n].y.parse);
    examples[n].x.si = si;
    /*
      int i;
      display_tree(stdout,examples[n].y.parse, si, 0); 
      printf("\n");
      for(i=0;i<examples[n].x.sentence.n;i++) {
        printf(" %s",si_index_string(si, examples[n].x.sentence.e[i]));
      }
      printf("\n");
    */
    if(examples[n].x.sentence.n <= sparm->maxsentlen)
      n++;
    else {
      free_label(examples[n].y);
      free(examples[n].x.sentence.e);
    }
    totsen++;
    if((struct_verbosity>=1) && ((n % 100) == 0)) { 
      printf("%ld..",n); 
      fflush(stdout);
    }
  }
  fclose(fp);
  if(struct_verbosity>=1) { 
      printf("(read %ld out of %ld)..",n,totsen); 
      fflush(stdout);
  }
  sample.n=n;
  sample.examples=examples;
  return(sample);
}

void        init_struct_model(SAMPLE sample, STRUCTMODEL *sm, 
			      STRUCT_LEARN_PARM *sparm)
{
  /* Initialize structmodel sm. The weight vector w does not need to be
     initialized, but you need to provide the maximum size of the
     feature space in sizePsi. This is the maximum number of different
     weights that can be learned. Later, the weight vector w will
     contain the learned weights for the model. */

  si_t     si;
  long     i;
  vihashl  localtree_ht = make_vihashl(1000);
  vihashl  weightid_ht = make_vihashl(1000);
  grammar  g;
  
  si=sample.examples[0].y.si;

  printf("Extracting grammar rules from training examples..."); fflush(stdout);
  for(i=0;i<sample.n;i++) {
    count_local_trees(sample.examples[i].y.parse, localtree_ht);
  }
  printf("done\n"); fflush(stdout);

  printf("Creating grammar..."); fflush(stdout);
  g = create_grammar(localtree_ht, weightid_ht, si);
  printf("done\n"); fflush(stdout);

  free_vihashl(localtree_ht);
  sm->grammar=g;
  sm->si=si;
  sm->weightid_ht=weightid_ht;
  sm->sizePsi=g.idMax+1;
  sm->sparm=sparm;
}

CONSTSET    init_struct_constraints(SAMPLE sample, STRUCTMODEL *sm, 
				    STRUCT_LEARN_PARM *sparm)
{
  /* Initializes the optimization problem. Typically, you do not need
     to change this function, since you want to start with an empty
     set of constraints. However, if for example you have constraints
     that certain weights need to be positive, you might put that in
     here. The constraints are represented as lhs[i]*w >= rhs[i]. lhs
     is an array of feature vectors, rhs is an array of doubles. m is
     the number of constraints. The function returns the initial
     set of constraints. */
  CONSTSET c;
  long     sizePsi=sm->sizePsi;
  long     i;
  WORD     words[2];

  if(1) { /* normal case: start with empty set of constraints */
    c.lhs=NULL;
    c.rhs=NULL;
    c.m=0;
  }
  else { /* add constraints so that all learned weights are
            positive. WARNING: Currently, they are positive only up to
            precision epsilon set by -e. */
    c.lhs=my_malloc(sizeof(DOC *)*sizePsi);
    c.rhs=my_malloc(sizeof(double)*sizePsi);
    for(i=0; i<sizePsi; i++) {
      words[0].wnum=i+1;
      words[0].weight=1.0;
      words[1].wnum=0;
      /* the following slackid is a hack. we will run into problems,
         if we have move than 1000000 slack sets (ie examples) */
      c.lhs[i]=create_example(i,0,1000000+i,1,create_svector(words,"",1.0));
      c.rhs[i]=0.0;
    }
  }
  return(c);
}

LABEL       classify_struct_example(PATTERN x, STRUCTMODEL *sm, 
				    STRUCT_LEARN_PARM *sparm)
{
  /* Finds the label yhat for pattern x that scores the highest
     according to the linear evaluation function in sm, especially the
     weights sm.w. The returned label is taken as the prediction of sm
     for the pattern x. The weights correspond to the features defined
     by psi() and range from index 1 to index sm->sizePsi. If the
     function cannot find a label, it shall return an empty label as
     recognized by the function empty_label(y). */
  LABEL   ybar;
  vindex  terms;
  grammar g;
  si_t    si;
  chart   c,c2;
  chart_cell	root_cell,root_cell2;
  FILE    *tracefp=NULL;
  FILE    *parsefp=NULL;
  double  *w;

  terms=&(x.sentence);
  w=sm->w;
  g=sm->grammar;
  si=sm->si;

  if (tracefp) {
    int i;
    fprintf(tracefp, "\nSentence:\n");
    for (i=0; i<terms->n; i++)
      fprintf(tracefp, " %s", si_index_string(si, terms->e[i]));
    fprintf(tracefp, "\n"); fflush(tracefp);
  }
     
  if (tracefp) { printf("Parsing sentence..."); fflush(stdout); }
  c = cky(*terms, NULL, sm, si, CKY_MAX);
  c2 = cky_wrap(*terms, NULL, sm, si); 
  if (tracefp) { printf("done\n"); fflush(stdout); }

  /* fetch best root node */

  if (tracefp) { printf("Fetching root..."); fflush(stdout); }
  root_cell = sihashcc_ref(CHART_ENTRY(c, 0, terms->n),
			   si_string_index(si, ROOT));
  root_cell2 = sihashcc_ref(CHART_ENTRY(c2, 0, terms->n),
			    si_string_index(si, ROOT));
  if (tracefp) { printf("done\n"); fflush(stdout); }

  ybar.si=si;
  if (root_cell) {
    if(fabs(root_cell->prob - root_cell2->prob) > 0.0000001) {
      printf("ERROR:\n");
      printf("prob=%g \t prob2=%g\n",root_cell->prob,root_cell2->prob);
      printf("Parse1: "); 
      write_tree(tracefp, bintree_tree(&root_cell->tree, si), si);
      printf("\nParse2: "); 
      write_tree(tracefp, bintree_tree(&root_cell2->tree, si), si);
      printf("\n"); 
      fflush(stdout); 
    }
    /* assert(fabs(root_cell->prob - root_cell2->prob) < 0.0000001); */

    if (tracefp) { printf("Getting parse tree..."); fflush(stdout); }
    ybar.parse = bintree_tree(&root_cell->tree, si);
    if (tracefp) { printf("done\n"); fflush(stdout); }
    
    double logprob = (double) root_cell->prob;
    if (parsefp) { fprintf(parsefp, "Prob = %g ", logprob); fflush(stdout); }
    ybar.prob=logprob;
  }
  else {
    ybar.parse=NULL;
    fprintf(stdout, "Failed to parse\n");
    if (parsefp)
      fprintf(parsefp, "parse_failure.\n");
  }

  chart_free(c, terms->n);			/* free the chart */
  chart_free(c2, terms->n);			/* free the chart */

  return(ybar);    
}

LABEL       find_most_violated_constraint_slackrescaling(PATTERN x, LABEL y, 
						     STRUCTMODEL *sm, 
						     STRUCT_LEARN_PARM *sparm)
{
  /* Finds the label ybar for pattern x that that is responsible for
     the most violated constraint for the slack rescaling
     formulation. It has to take into account the scoring function in
     sm, especially the weights sm.w, as well as the loss
     function. The weights in sm.w correspond to the features defined
     by psi() and range from index 1 to index sm->sizePsi. Most simple
     is the case of the zero/one loss function. For the zero/one loss,
     this function should return the highest scoring label ybar, if
     ybar is unequal y; if it is equal to the correct label y, then
     the function shall return the second highest scoring label. If
     the function cannot find a label, it shall return an empty label
     as recognized by the function empty_label(y). */
  LABEL   ybar;
  vindex  terms;
  grammar g;
  si_t    si;
  chart   c,c2;
  chart_cell	root_cell,root_cell2;
  FILE    *tracefp=NULL;
  FILE    *parsefp=NULL;
  double  *w;
  double  loss;

  terms=&(x.sentence);
  w=sm->w;
  g=sm->grammar;
  si=sm->si;

  if (tracefp) {
    int i;
    fprintf(tracefp, "\nSentence:\n");
    for (i=0; i<terms->n; i++)
      fprintf(tracefp, " %s", si_index_string(si, terms->e[i]));
    fprintf(tracefp, "\n"); fflush(tracefp);
  }
     
  if (tracefp) { printf("Parsing sentence..."); fflush(stdout); }

  if(sparm->loss_function == 0) { 
    c = cky(*terms, NULL, sm, si, CKY_MAX); 
    /* c2 = cky_wrap(*terms, y.parse, sm, si); */
    c2=c;
    loss=1.0; 
  }
  else  if(sparm->loss_function >= 1) { /* type 3: stratified losses */
    c = cky_maxloss(*terms, x, y, sm, sparm, &loss); 
    c2=c;
  }
  else{
    c=NULL;
    c2=NULL;
    assert(NULL);
  }
  if (tracefp) { printf("done\n"); fflush(stdout); }

  /* fetch best root node */

  if (tracefp) { printf("Fetching root..."); fflush(stdout); }
  root_cell = sihashcc_ref(CHART_ENTRY(c, 0, terms->n),
			   si_string_index(si, ROOT));
  root_cell2 = sihashcc_ref(CHART_ENTRY(c2, 0, terms->n),
			    si_string_index(si, ROOT));
  if (tracefp) { printf("done\n"); fflush(stdout); }

  ybar.si=si;
  ybar.loss = loss;
  if (root_cell) {
    assert(fabs(root_cell->prob - root_cell2->prob) < 0.0000001); 

    if (tracefp) { printf("Getting parse tree..."); fflush(stdout); }
    ybar.parse = bintree_tree(&root_cell->tree, si);
    if (tracefp) { printf("done\n"); fflush(stdout); }
    
    double logprob = (double) root_cell->prob;
    if (parsefp) { fprintf(parsefp, "Prob = %g ", logprob); fflush(stdout); }
    ybar.prob=logprob;
    
    { /* sanity check */
      SVECTOR *v=psi(x,ybar,sm,sparm);
      DOC *ex=create_example(-1,-1,-1,1,v);
      double logprobpsi=classify_example(sm->svm_model,ex);
      free_example(ex,0);
      if (parsefp) { fprintf(parsefp, "ProbPsi = %g ", logprobpsi); fflush(stdout); }
      if(fabs(logprob - logprobpsi) > 0.000001) 
	printf(" psi-cky mismatch\n"); 
      assert(fabs(logprob - logprobpsi) < 0.000001); 
      free_svector(v);
      if(root_cell->present2) {
	LABEL ybar2;
	ybar2.parse = bintree_tree(&root_cell->tree2, si);
	assert(root_cell->prob2 <= root_cell->prob);
	double logprob = (double) root_cell->prob2;
	v=psi(x,ybar2,sm,sparm);
	ex=create_example(-1,-1,-1,1,v);
	logprobpsi=classify_example(sm->svm_model,ex);
	free_example(ex,0);
	if (parsefp) { fprintf(parsefp, " > %g ", logprobpsi); fflush(stdout); }
	if(fabs(logprob - logprobpsi) > 0.0000001)
	  printf(" psi-cky2 mismatch\n");
	if(tree_eq(ybar.parse,ybar2.parse) || (fabs(logprob - logprobpsi) > 0.0000001)) {
	  printf("\n   prob1=%f      prob2=%f\n",root_cell->prob,root_cell->prob2);
	  /* printf("\npsiprob1=%f   psiprob2=%f %f %f\n",sprod_ns(sm->grammar.weight, psi(x,ybar,sm,sparm)),sprod_ns(sm->grammar.weight, psi(x,ybar2,sm,sparm)),logprob,logprobpsi); */
	  write_bintree(stdout,&root_cell->tree, si);printf("\n");
	  write_bintree(stdout,&root_cell->tree2, si);printf("\n");
	  write_tree(stdout,ybar.parse, si);printf("\n");
	  write_tree(stdout,ybar2.parse, si);printf("\n");
	}
	assert(!tree_eq(ybar.parse,ybar2.parse)); 
	assert(fabs(logprob - logprobpsi) < 0.0000001); 
	free_svector(v);
	free_tree(ybar2.parse);
      }
    }

    if(((sparm->loss_function >= 1)) && (ybar.loss == 0)) {
      free_label(ybar);
      ybar.parse=NULL;
    }

    if(tree_eq(ybar.parse,y.parse) && (sparm->loss_function == 0)) { /* return second best parse */
      if (tracefp) { 
	printf("Best parse is correct:\n"); 
	printf("Correct: "); 
	write_tree(tracefp, y.parse, si);
	printf("\nBest   : "); 
	write_tree(tracefp, ybar.parse, si);
	printf("\n"); 
	fflush(stdout); 
      }
      free_tree(ybar.parse);

      if(root_cell->present2) {
	ybar.parse = bintree_tree(&root_cell->tree2, si);
	if (tracefp && ybar.parse) { 
	  printf("2ndBest: "); 
	  write_tree(tracefp, ybar.parse, si);
	  printf("\n"); 
	  fflush(stdout); 
	}
      }
      else {
	if (tracefp) { 
	  printf("No 2nd best parse found!\n"); 
	  fflush(stdout); 
	}
	ybar.parse=NULL;
      }
    }
    if (parsefp && ybar.parse)
      write_tree(parsefp, ybar.parse, si);
  }
  else {
    ybar.parse=NULL;
    fprintf(stderr, "Failed to parse\n");
    if (parsefp)
      fprintf(parsefp, "parse_failure.\n");
  }

  if(ybar.parse)  
    assert(!tree_eq(ybar.parse,y.parse)); 

  chart_free(c, terms->n);			/* free the chart */
  if(c != c2) chart_free(c2, terms->n);		/* free the chart */

  return(ybar);
}

LABEL       find_most_violated_constraint_marginrescaling(PATTERN x, LABEL y, 
						     STRUCTMODEL *sm, 
						     STRUCT_LEARN_PARM *sparm)
{
  /* Finds the label ybar for pattern x that that is responsible for
     the most violated constraint for the margin rescaling
     formulation. It has to take into account the scoring function in
     sm, especially the weights sm.w, as well as the loss
     function. The weights in sm.w correspond to the features defined
     by psi() and range from index 1 to index sm->sizePsi. Most simple
     is the case of the zero/one loss function. For the zero/one loss,
     this function should return the highest scoring label ybar, if
     ybar is unequal y; if it is equal to the correct label y, then
     the function shall return the second highest scoring label. If
     the function cannot find a label, it shall return an empty label
     as recognized by the function empty_label(y). */
  LABEL ybar;

  ybar=find_most_violated_constraint_slackrescaling(x,y,sm,sparm);

  return(ybar);
}

int         empty_label(LABEL y)
{
  /* Returns true, if y is an empty label. An empty label might be
     returned by find_most_violated_constraint_???(x, y, sm) if there
     is no incorrect label that can be found for x, or if it is unable
     to label x at all */
  return(y.parse == NULL);
}

SVECTOR     *psi(PATTERN x, LABEL y, STRUCTMODEL *sm,
		 STRUCT_LEARN_PARM *sparm)
{
  /* Returns a feature vector describing the match between pattern x and
     label y. The feature vector is returned as an SVECTOR
     (i.e. pairs <featurenumber:featurevalue>), where the last pair has
     featurenumber 0 as a terminator. Featurenumbers start with 1 and end with
     sizePsi. This feature vector determines the linear evaluation
     function that is used to score labels. There will be one weight in
     sm.w for each feature. Note that psi has to match
     find_most_violated_constraint_???(x, y, sm) and vice versa. In
     particular, find_most_violated_constraint_???(x, y, sm) finds that
     ybar!=y that maximizes psi(x,ybar,sm)*sm.w (where * is the inner
     vector product) and the appropriate function of the loss.  */
  SVECTOR *psi;
  bintree btree;
  struct vindex  terms;
  size_t  rpos;

  terms=tree_terms(y.parse);
  btree=right_binarize(y.parse,sm->si);
  psi=collect_phi(btree,sm,0,&rpos,0,terms.n);
  free(terms.e);
  free_bintree(btree);

  return(psi);
}

double      loss(LABEL y, LABEL ybar, STRUCT_LEARN_PARM *sparm)
{
  /* loss for correct label y and predicted label ybar. The loss for
     y==ybar has to be zero. sparm->loss_function is set with the -l option. */
  if(sparm->loss_function == 0) { /* type 0 loss: 0/1 loss */
                                  /* return 0, if y==ybar. return 1 else */
    if(tree_eq(ybar.parse,y.parse))
      return(0);
    else 
      return(1);
  }
  else if((sparm->loss_function == 1)) {  
    /* type 1 loss: 1-fone */
    tree t = remove_parent_annotation(y.parse, y.si); 
    struct ledges *test_ledges = tree_ledges(t);   
    free_tree(t);
    t = remove_parent_annotation(ybar.parse, ybar.si); 
    struct ledges *parse_ledges = tree_ledges(t);   
    free_tree(t);
    int common_bracket_count = common_ledge_count(test_ledges, parse_ledges);
    double loss=1.0-fone(common_bracket_count,test_ledges->n,parse_ledges->n);
    free_ledges(test_ledges);
    free_ledges(parse_ledges);

    /* printf("loss: %g ybar.loss=%g\n",loss,ybar.loss); */
    /* assert(fabs(loss - ybar.loss) < 0.00000001); */
    return(loss);  
  }
  else {
    /* Put your code for different loss functions here. But then
       find_most_violated_constraint_???(x, y, sm) has to return the
       highest scoring label with the largest loss. */
    assert(0);
    return(0);
  }
}

void        print_struct_learning_stats(SAMPLE sample, STRUCTMODEL *sm,
					CONSTSET cset, double *alpha, 
					STRUCT_LEARN_PARM *sparm)
{
  /* This function is called after training and allows final touches to
     the model sm. But primarly it allows computing and printing any
     kind of statistic (e.g. training error) you might want. */
  long i,correct=0;
  double valy,valybar;
  LABEL ybar;
  EXAMPLE *ex;
  DOC *extemp;
  SVECTOR *fy,*fybar;
  long n;
  double cumloss=0;

  n=sample.n;
  ex=sample.examples;

  if(struct_verbosity >= 1)
    printf("Classifying training examples");

  /* classify the training examples */
  for(i=0; i<n;i++) { 
    /* find best parse */
    ybar=classify_struct_example(ex[i].x,sm,sparm);
    assert(ybar.parse);

    /* compute scores for sanity check */
    fy=psi(ex[i].x,ex[i].y,sm,sparm);
    extemp=create_example(-1,-1,-1,1,fy);
    valy=classify_example(sm->svm_model,extemp);
    free_example(extemp,1);
    fybar=psi(ex[i].x,ybar,sm,sparm);
    extemp=create_example(-1,-1,-1,1,fybar);
    valybar=classify_example(sm->svm_model,extemp);
    free_example(extemp,1);

    if(tree_eq(ybar.parse,ex[i].y.parse)) {
      correct++;
      printf("+");fflush(stdout);
      assert(valy >= valybar);
    }
    else {
      printf("-");fflush(stdout);
      cumloss+=loss(ex[i].y,ybar,sparm);
      assert(valy <= valybar);
    }

    /*
    fprintf(stdout,"Prob = %g ", ybar.prob); fflush(stdout);
    fprintf(stdout,"Xi = %g = %g ", xi, sm->w[sm->sizePsi+1+i]/sqrt(2*sparm->C)); fflush(stdout);
    if(sparm->parent_annotation) {
      tree t = remove_parent_annotation(ybar.parse, sm->si);
      free_tree(ybar.parse);
      ybar.parse=t;
    }
    write_prolog_tree(stdout, ybar.parse, sm->si);
    */

    free_label(ybar);
  }
  printf("done\n");
  if(struct_verbosity>=1) {
    printf("Number of correct parses: %i out of %i (%.2f%%)\n",
	   (int)correct,(int)n, 100.0*correct/n);
    printf("Average training loss: %.4f\n", cumloss/n);
  }
}

void        print_struct_testing_stats(SAMPLE sample, STRUCTMODEL *sm,
				       STRUCT_LEARN_PARM *sparm, 
				       STRUCT_TEST_STATS *teststats)
{
  /* This function is called after making all test predictions in
     svm_struct_classify and allows computing and printing any kind of
     evaluation (e.g. precision/recall) you might want. You can use
     the function eval_prediction to accumulate the necessary
     statistics for each prediction. */
  if(verbosity >= 1) {
    printf("Fraction of parsed sentences: %.2f%% (%ld/%d)\n",
	   100.0*teststats->parsed_sentences/sample.n,
	   teststats->parsed_sentences,sample.n);
    printf("Labelled bracket precision: %.2f%% (%ld/%ld)\n", 
	    100.0*teststats->common_bracket_sum/teststats->parse_bracket_sum,
	    teststats->common_bracket_sum, teststats->parse_bracket_sum);
    printf("Labelled bracket recall: %.2f%% (%ld/%ld)\n", 
	    100.0*teststats->common_bracket_sum/teststats->test_bracket_sum,
	    teststats->common_bracket_sum, teststats->test_bracket_sum);
    printf("Labelled bracket F1: %.2f%%\n", 
	   100.0*fone(teststats->common_bracket_sum,
		      teststats->test_bracket_sum, 
		      teststats->parse_bracket_sum));
   }
}

void        eval_prediction(long exnum, EXAMPLE ex, LABEL ypred, 
			    STRUCTMODEL *sm, STRUCT_LEARN_PARM *sparm, 
			    STRUCT_TEST_STATS *teststats)
{
  /* This function allows you to accumlate statistic for how well the
     predicition matches the labeled example. It is called from
     svm_struct_classify. See also the function
     print_struct_testing_stats. */
  if(exnum == 0) { /* this is the first time the function is
		      called. So initialize the teststats */
    teststats->parsed_sentences=0;
    teststats->test_bracket_sum=0;
    teststats->parse_bracket_sum=0;
    teststats->common_bracket_sum=0;
  }

  if(!empty_label(ypred)) {
    teststats->parsed_sentences++;
    tree t = remove_parent_annotation(ex.y.parse, ex.y.si); 
    struct ledges *test_ledges=tree_ledges(t);
    free_tree(t);
    t = remove_parent_annotation(ypred.parse, ypred.si); 
    struct ledges *parse_ledges=tree_ledges(t);
    free_tree(t);
    teststats->common_bracket_sum+=common_ledge_count(test_ledges, parse_ledges);
    teststats->test_bracket_sum+=test_ledges->n;
    teststats->parse_bracket_sum+=parse_ledges->n;
    free_ledges(test_ledges);
    free_ledges(parse_ledges);
  }
}

void        write_struct_model(char *file, STRUCTMODEL *sm, 
			       STRUCT_LEARN_PARM *sparm)
{
  vihashlit hit;
  vindex    vi;
  long      weightid,i;
  FILE      *fp;
  char      file_svm[500],file_grammar[500];

  /* First write the SVM model */
  strcpy(file_svm,file);
  strcat(file_svm,".svm");
  write_model(file_svm,sm->svm_model);

  /* Then write the grammar */
  strcpy(file_grammar,file);
  strcat(file_grammar,".grammar");
  if ((fp = fopen (file_grammar, "w")) == NULL)
    { perror (file_grammar); exit (1); }
  fprintf(fp,"%ld # number of attributes in weight vector\n",sm->sizePsi); 
  fprintf(fp,"%d # loss function\n",sparm->loss_function);
  fprintf(fp,"%d # use parent annotation\n",sparm->parent_annotation);
  fprintf(fp,"%d # maximum sentence length\n",sparm->maxsentlen);
  fprintf(fp,"%d # use border features\n",sparm->feat_borders);
  fprintf(fp,"%d # use span length as feature\n",
	  sparm->feat_parent_span_length);
  fprintf(fp,"%d # use children span length as features\n",
	  sparm->feat_children_span_length);
  fprintf(fp,"%d # use difference in children span lengths as feature\n",
	  sparm->feat_diff_children_length);
  for (hit = vihashlit_init(sm->weightid_ht); vihashlit_ok(hit); hit = vihashlit_next(hit)) {
    vi=hit.key;
    weightid=vihashl_ref(sm->weightid_ht,vi);
    assert(vi->n);
    fprintf(fp,"%ld %s " REWRITES "",weightid,si_index_string(sm->si, vi->e[0]));
    for(i=1;i<vi->n;i++) {
      fprintf(fp," %s",si_index_string(sm->si, vi->e[i]));
    }
    fprintf(fp,"\n");
  }
  fclose(fp);
}

STRUCTMODEL read_struct_model(char *file, STRUCT_LEARN_PARM *sparm)
{
  /* Reads structural model sm from file file. This function is used
     only in the prediction module, not in the learning module. */
  STRUCTMODEL sm;
  vindex    vi;
  long      i,id;
  FILE      *fp;
  char      file_svm[500],file_grammar[500];
  sihashbrsit	bhit;
  sihashursit	uhit;

  /* First write the SVM model */
  strcpy(file_svm,file);
  strcat(file_svm,".svm");
  sm.svm_model=read_model(file_svm);

  /* Then read the grammar */
  strcpy(file_grammar,file);
  strcat(file_grammar,".grammar");
  if ((fp = fopen (file_grammar, "r")) == NULL)
    { perror (file_grammar); exit (1); }
  fscanf(fp,"%ld%*[^\n]\n",&sm.sizePsi); 
  fscanf(fp,"%d%*[^\n]\n", &sparm->loss_function);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->parent_annotation);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->maxsentlen);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->feat_borders);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->feat_parent_span_length);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->feat_children_span_length);  
  fscanf(fp,"%d%*[^\n]\n", &sparm->feat_diff_children_length);  
  sm.si=make_si(1024);
  sm.grammar=read_grammar(fp,sm.si);
  sm.grammar.idMax=sm.sizePsi;
  sm.sparm=sparm;
  fclose(fp);

  /* Reconstruct the hashtable for mapping rules to feature numbers */
  sm.weightid_ht = make_vihashl(1000);
  for (bhit=sihashbrsit_init(sm.grammar.brs); sihashbrsit_ok(bhit); bhit=sihashbrsit_next(bhit)) 
    for (i=0; i<bhit.value.n; i++) {
      vi=make_vindex(4); 
      vi->n=3;
      vi->e[0]=bhit.value.e[i]->parent;
      vi->e[1]=bhit.value.e[i]->left;
      vi->e[2]=bhit.value.e[i]->right;
      id=(long)(bhit.value.e[i]->prob+0.1);
      vihashl_set(sm.weightid_ht,vi,id);	
      bhit.value.e[i]->weightid=id;
      vindex_free(vi);
    }
  for (uhit=sihashursit_init(sm.grammar.urs); sihashursit_ok(uhit); uhit=sihashursit_next(uhit)) 
    for (i=0; i<uhit.value.n; i++) {
      vi=make_vindex(4); 
      vi->n=2;
      vi->e[0]=uhit.value.e[i]->parent;
      vi->e[1]=uhit.value.e[i]->child;
      id=(long)(uhit.value.e[i]->prob+0.1);
      vihashl_set(sm.weightid_ht,vi,id);	
      uhit.value.e[i]->weightid=id;
      vindex_free(vi);
    }

  sparm->si=sm.si;
  return(sm);
}

void        write_label(FILE *fp, LABEL y)
{
  /* Writes label y to file handle fp. */
  tree t = remove_parent_annotation(y.parse, y.si); 
  if(y.parse)
    write_tree(fp, t, y.si);
  else 
    fprintf(fp,"NULL");
  fprintf(fp,"\n");
  free_tree(t);
} 

void        free_pattern(PATTERN x) {
  /* Frees the memory of x. */
  free(x.sentence.e);
}

void        free_label(LABEL y) {
  /* Frees the memory of y. */
  if(y.parse) free_tree(y.parse);
}

void        free_struct_model(STRUCTMODEL sm) 
{
  /* Frees the memory of model. */
  /* if(sm.w) free(sm.w); */ /* this is free'd in free_model */
  if(sm.svm_model) free_model(sm.svm_model,1);
  /* add free calls for user defined data here */
  free_grammar(sm.grammar);
  free_vihashl(sm.weightid_ht);
  si_free(sm.si);
}

void        free_struct_sample(SAMPLE s)
{
  /* Frees the memory of sample s. */
  int i;
  for(i=0;i<s.n;i++) { 
    free_pattern(s.examples[i].x);
    free_label(s.examples[i].y);
  }
  free(s.examples);
}

void        print_struct_help()
{
  /* Prints a help text that is appended to the common help text of
     svm_struct_learn. */
  printf("         --s [1..]   -> Maximum length of sentences to use (default 10)\n");
  printf("         --p {0,1}   -> Use parent annotation for rules (default 0)\n");
  printf("         --a {0,1}   -> Use borders as feature (default 0)\n");
  printf("         --b {0,1}   -> Use length of parent span as feature (default 0)\n");
  printf("         --c {0,1}   -> Use length of childrens spans as features (default 0)\n");
  printf("         --d {0,1}   -> Use difference between children spans as feature\n");
  printf("                        (default 0)\n");
  printf("         -l {0,1}    NOTE: Added loss type '1' in addition to the 0/1-loss '0'.\n");
  printf("                     Loss type '1' uses (1-F_1) as the loss as in [2].\n");
  printf("\n");
  printf("The CKY parser is based on the implementation of Mark Johnson described in:\n");
  printf("   Johnson, M. (1999). PCFG models of linguistic tree representations.\n");
  printf("   Computational Linguistics.\n");
}
 
void         parse_struct_parameters(STRUCT_LEARN_PARM *sparm)
{
  /* Parses the command line parameters that start with -- */
  int i;

  sparm->maxsentlen=10;
  sparm->parent_annotation=0;
  sparm->feat_borders=0;
  sparm->feat_parent_span_length=0;
  sparm->feat_children_span_length=0;
  sparm->feat_diff_children_length=0;
  sparm->si=NULL;

  for(i=0;(i<sparm->custom_argc) && ((sparm->custom_argv[i])[0] == '-');i++) {
    switch ((sparm->custom_argv[i])[2]) 
      { 
      case 's': i++; sparm->maxsentlen=atol(sparm->custom_argv[i]); break;
      case 'p': i++; sparm->parent_annotation=atol(sparm->custom_argv[i]); break;
      case 'a': i++; sparm->feat_borders=atol(sparm->custom_argv[i]); break;
      case 'b': i++; sparm->feat_parent_span_length=atol(sparm->custom_argv[i]); break;
      case 'c': i++; sparm->feat_children_span_length=atol(sparm->custom_argv[i]); break;
      case 'd': i++; sparm->feat_diff_children_length=atol(sparm->custom_argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",sparm->custom_argv[i]);
	       exit(0);
      }
  }
}


/* ---------------------------------------------------------------- */
  
SVECTOR *collect_phi(bintree parse, STRUCTMODEL *sm, 
		     size_t lpos, size_t *rpos, size_t start, size_t end)
{
  SVECTOR *vecleft=NULL, *vecright=NULL, *sum, *vec;
  long    weightid, mpos=0;
  vindex  vi;

  vi=make_vindex(4);
  vi->e[0]=parse->label;
  vi->n=1;

  if(parse->left) {   /* construct binary rule from tree */
    vi->e[1]=parse->left->label;
    vi->n=2;
    vecleft=collect_phi(parse->left,sm,lpos,rpos,start,end);
    if(parse->right) {
      vi->e[2]=parse->right->label;
      vi->n=3;
      mpos=*rpos;
      vecright=collect_phi(parse->right,sm,mpos,rpos,start,end);
    }
  }
  else {
    *rpos=lpos+1;
  }

  if(parse->left && parse->right) { /* binary rule */
    weightid=vihashl_ref(sm->weightid_ht,vi);
    assert(weightid);
    vec=phi_brule(weightid,parse->left,parse->right,sm,lpos,mpos,
		  *rpos,start,end);
    vec->next=vecleft;                /* prepend vec to vec of left tree */
    append_svector_list(vec,vecright);/* concatenate all for preorder list */
    sum=vec;
  }
  else if(parse->left) {           /* unary rule */
    weightid=vihashl_ref(sm->weightid_ht,vi);
    assert(weightid);
    vec=phi_urule(weightid,parse->left,sm,lpos,*rpos,start,end);
    vec->next=vecleft;                /* prepend vec to vec of left tree */
    sum=vec;
  }
  else {                           /* terminal */
    sum=NULL;
  }

  vindex_free(vi);

  return(sum);
}


SVECTOR *phi_urule(long weightid, bintree child, STRUCTMODEL *sm, 
		   int lpos, int rpos, int start, int end)
{
  WORD    *feat;
  SVECTOR *vec;
  long    basefnum;
  long    pos;
  WORD    empty[2];

  assert(weightid != 0);
  if(weightid <= 0) {
    empty[0].wnum=0;
    return(create_svector(empty,"",1.0));
  }

  basefnum=weightid;
  feat=(WORD *)MALLOC(sizeof(WORD)*MAXFEAT);
  feat[0].wnum=0;
  pos=0;
  
  /* base feature */
  pos=add_feature(feat,pos,basefnum+0,1.0);

  /* borders beginning or end */
  if(sm->sparm->feat_borders) {
    if(lpos == start) 
      pos=add_feature(feat,pos,basefnum+1,1.0);
    if(rpos == end) 
      pos=add_feature(feat,pos,basefnum+2,1.0);
  }

  /* length of span of whole rule */
  if((sm->sparm->feat_parent_span_length) 
     || (sm->sparm->feat_children_span_length))
     pos=encode_number(feat,pos,basefnum+3,rpos-lpos,1.0,2,3,4,5,7,9);

  assert(pos<MAXFEAT);

  vec=create_svector(feat,"",1.0);
  free(feat);

  return(vec);
}


SVECTOR *phi_brule(long weightid, bintree left, bintree right, STRUCTMODEL *sm,
		   int lpos, int mpos, int rpos, int start, int end)
{
  WORD *feat;
  SVECTOR *vec;
  long basefnum;
  long pos;
  WORD    empty[2];

  assert(weightid != 0);
  if(weightid <= 0) {
    empty[0].wnum=0;
    return(create_svector(empty,"",1.0));
  }

  basefnum=weightid;
  feat=(WORD *)MALLOC(sizeof(WORD)*(MAXFEAT+1));
  feat[0].wnum=0;
  pos=0;
  
  /* base feature */
  pos=add_feature(feat,pos,basefnum+0,1.0);

  /* borders beginning or end */
  if(sm->sparm->feat_borders) {
    if(lpos == start) 
      pos=add_feature(feat,pos,basefnum+1,1.0);
    if(rpos == end) 
      pos=add_feature(feat,pos,basefnum+2,1.0);
  }

  /* length of span of whole rule */
  if(sm->sparm->feat_parent_span_length) 
     pos=encode_number(feat,pos,basefnum+3,rpos-lpos,1.0,2,3,4,5,7,9);

  /* length of span of each child */
  if(sm->sparm->feat_children_span_length) {
     pos=encode_number(feat,pos,basefnum+9,rpos-mpos,1.0,2,3,4,5,7,9);
     pos=encode_number(feat,pos,basefnum+15,mpos-lpos,1.0,2,3,4,5,7,9);
  }

  /* difference between left and right child */
  if(sm->sparm->feat_diff_children_length) 
     pos=encode_number(feat,pos,basefnum+21,(rpos-mpos)-(mpos-lpos),1.0,-7,-3,-1,1,3,7);

  assert(pos<MAXFEAT);

  vec=create_svector(feat,"",1.0);
  free(feat);

  return(vec);
}


int add_feature(WORD *feat, int pos, long fnum, long weight)
     /* add feature fnum with value weight to end of feature list */ 
{
  /* find end of feature list */
  for(pos=0;feat[pos].wnum>0;pos++);
  feat[pos].wnum=fnum;
  feat[pos].weight=weight;
  pos++;
  feat[pos].wnum=0;
  return(pos);
}


int encode_number(WORD *feat,int pos,long basefnum,long number,double weight,
		       long a, long b, long c, long d, long e, long f) 
     /* encodes a number as a vector of 6 featurs by thresholding it
	against a,b,c,d,e,f */
{
  /* find end of feature list */
  for(pos=0;feat[pos].wnum>0;pos++);

  if(number <= a) 
    pos=add_feature(feat,pos,basefnum+0,weight);
  if(number <= b) 
    pos=add_feature(feat,pos,basefnum+1,weight);
  if(number <= c) 
    pos=add_feature(feat,pos,basefnum+2,weight);
  if(number <= d) 
    pos=add_feature(feat,pos,basefnum+3,weight);
  if(number <= e) 
    pos=add_feature(feat,pos,basefnum+4,weight);
  if(number <= f) 
    pos=add_feature(feat,pos,basefnum+5,weight);

  return(pos);
}


double urule_value(urule rule, bintree child, STRUCTMODEL *sm, 
		   int lpos, int rpos, int start, int end)
{
  double val;
  SVECTOR *vec;
  DOC *ex;

  assert(rule->weightid >= -1);
  assert(rule->weightid != 0);
  assert(rule->weightid < sm->sizePsi);
  /* printf("urule: %ld\n",rule->weightid); */
  if(rule->weightid == -1) {
    return(0.0);
  }
  else {
    vec=phi_urule(rule->weightid,child,sm,lpos,rpos,start,end);
    ex=create_example(-1,-1,-1,1,vec);
    val=classify_example(sm->svm_model,ex);
    free_example(ex,1);
    return(val);
  }
}

double brule_value(brule rule, bintree left, bintree right, STRUCTMODEL *sm, 
		   int lpos, int mpos, int rpos, int start, int end)
{
  double val;
  SVECTOR *vec;
  DOC *ex;

  assert(rule->weightid >= -1);
  assert(rule->weightid != 0);
  assert(rule->weightid < sm->sizePsi);
  /* printf("brule: %ld\n",rule->weightid); */
  if(rule->weightid == -1) {
    return(0.0);
  }
  else {
    vec=phi_brule(rule->weightid,left,right,sm,lpos,mpos,rpos,start,end);
    ex=create_example(-1,-1,-1,1,vec);
    val=classify_example(sm->svm_model,ex);
    free_example(ex,1);
    return(val);
  }
}

void count_local_trees(const tree t, vihashl localtree_ht)
{
  si_index      e[MAXRHS];
  struct vindex vi = {0, e};
  tree	        p;

  if (!t || !t->subtrees) return;

  vi.n = 0;
  vi.e[vi.n++] = t->label;

  for(p = t->subtrees; p; p=p->sibling) {
    assert(vi.n < MAXRHS);
    vi.e[vi.n++] = p->label;
    count_local_trees(p, localtree_ht);
  }

  vihashl_inc(localtree_ht, &vi, 1);	/* save this list */
}

void write_local_trees(FILE *fp,const vihashl localtree_ht, si_t si)
{
  vihashlit hit;
  vindex  vi;
  long    count;
  size_t  i;
  char    *string;

  for (hit = vihashlit_init(localtree_ht); vihashlit_ok(hit); hit = vihashlit_next(hit)) {
    vi = (vindex) hit.key;
    assert(vi->n > 0);
    assert(vi->n <= MAXRHS);
    count = hit.value;
    string =  si_index_string(si, vi->e[0]);
    assert(string);
    fprintf(fp,"%ld\t%s " REWRITES, count, string);
    
    for (i=1; i<vi->n; i++) {
      string = si_index_string(si, vi->e[i]);
      assert(string);
      fprintf(fp," %s", string);
    }

    fprintf(fp,"\n");
  }}

int tree_eq(const tree t1, const tree t2)
{
  tree	        p1,p2;

  if ((!t1) && (!t2)) return(1);
  if ((!t1) || (!t2)) return(0);
  if(t1->label == t2->label) {
    for(p1 = t1->subtrees,p2 = t2->subtrees; p1 || p2; p1=p1->sibling,p2=p2->sibling) {
      if(!tree_eq(p1,p2))
	return(0);
    }
  }
  else 
    return(0);

  return(1);
}



/* ------- parser -------- */

chart
chart_make(size_t n)
{
  size_t  i, nn = CHART_SIZE(n);
  chart   c = MALLOC(nn*sizeof(sihashcc));
  
  for (i=0; i<nn; i++) 
    c[i] = NULL;	/* chart cell will be constructed in apply_unary() */ 
  
  return c;
}

void
chart_free(chart c, size_t n)
{
  size_t i;

  for (i=0; i<CHART_SIZE(n); i++)
    free_sihashcc(c[i]);

  FREE(c);
}

void
chart_entry_display(FILE *fp, sihashcc chart_entry, si_t si)
{
  sihashccit hit;
  tree t;

  for (hit=sihashccit_init(chart_entry); sihashccit_ok(hit); 
       hit = sihashccit_next(hit)) {
    fprintf(fp, "\n %s: %g \t", si_index_string(si, hit.key),
	    (double) hit.value->prob);
    t = bintree_tree(&hit.value->tree, si);
    write_tree(fp, t, si);
    free_tree(t);
    if(hit.value->present2) {
      fprintf(fp, "\t : %g \t", (double) hit.value->prob2);
      t = bintree_tree(&hit.value->tree2, si);
      write_tree(fp, t, si);
      free_tree(t);
    }
  }
  fprintf(fp, "\n---\n");
}
  
int is_urule_cycle(bintree tree, si_index label, si_t si)
{
  if(!tree)                      /* is empty tree */
    return(0);
  else if(remove_parent_from_label(tree->label, si) 
	  == remove_parent_from_label(label,si))  
                                 /* already member of unary chain */
    return(1);
  else if(tree->right)           /* not produced by a unary rule */
    return(0);
  else {
    return(is_urule_cycle(tree->left,label,si));
  }
}

int
add_edge_zeroone(sihashcc chart_entry, si_index label, 
		 bintree left, bintree right,
		 FLOAT prob, int add_mode, si_t si)
{
  if(add_mode == NO_ADD_NEW_SYMBOLS) 
    if(!sihashcc_ref(chart_entry,label))
      return(0);

  chart_cell *cp = sihashcc_valuep(chart_entry, label);
  chart_cell cc = *cp;

  if (cc == NULL) {  /* construct a new chart entry */
    *cp = make_chart_cell(label, left, right, prob, 1, 0, CKY_MAX);
    return(1);
  }  
  
  /* we're dealing with an old chart entry */
  
  /* avoid cyclic chains of unary rules */
  if((!right)                /* unary rule is applied */
     && is_urule_cycle(left,label,si)) { 
    /* printf("found cycle with %ld\n",label); fflush(stdout); */
    return(0); 
  }

  assert(cc->tree.label==label);
  assert(cc->tree2.label==label);
  assert((cc->prob2 <= cc->prob) || (!cc->present2));

  /* check for duplicate rule application and side effects of unary rules */
  if((cc->tree.left == left) && (cc->tree.right == right)) {
    if(cc->prob < prob) { /* update prob of first */
      cc->prob=prob;
      return(1);
    }
    else 
      return(0);
  }

  if ((cc->prob2 >= prob) && (cc->present2))
    return(0);      /* second best chart cell entry is better than current */

  if (cc->prob > prob) { /* better than second best, but worse than best */
    cc->present2=1;
    cc->tree2.left = left;
    cc->tree2.right = right;
    cc->prob2 = prob;
    cc->nalt2 = 1;
    return(1);
    /* WARNING: Ties are not broken at random for second best */
  }

  if (cc->prob < prob) {  /* new entry is better than the best old one */
    /* move former best down to second best */
    cc->present2=1;
    cc->tree2.left = cc->tree.left;
    cc->tree2.right = cc->tree.right;
    cc->prob2 = cc->prob;
    cc->nalt2 = 1;
    /* make new one the best */
    cc->tree.left = left;
    cc->tree.right = right;
    cc->prob = prob;
    cc->nalt = 1;
    return(1);
  }

  /* old best and new entry have same probability */

  assert(cc->nalt<UINT_MAX);

  if (rand() > RAND_MAX/(++(cc->nalt))) {
    /* make new one the second best */
    cc->present2=1;
    cc->tree2.left = left;
    cc->tree2.right = right;
    cc->prob2 = prob;
    cc->nalt2 = 1;
    return(1);
  }

  /* make new one the best */
  /* move former best down to second best */
  cc->present2=1;
  cc->tree2.left = cc->tree.left;
  cc->tree2.right = cc->tree.right;
  cc->prob2 = cc->prob;
  cc->nalt2 = 1;
  /* make new one the best */
  cc->tree.left = left;
  cc->tree.right = right;
  cc->prob = prob;
  return(1);
}

int
add_edge_stratified(chart_cell cc, si_index label, bintree left, 
		    bintree right, FLOAT prob, int correct, int size, si_t si)
{
  long k;
  strat_cell *sp, ss;

  assert(label == cc->tree.label);

  /* avoid cyclic chains of unary rules */
  if((!right)                /* unary rule is applied */
     && is_urule_cycle(left,label,si)) { 
    /* printf("found cycle with %ld\n",label); fflush(stdout); */
    return(0); 
  }

  k=correct*1000+size;
  sp = strat_hash_valuep(cc->strat_hash, k);
  ss = *sp;
  if(ss == NULL) { /* new entry */
    *sp=MALLOC(sizeof(struct strat_cell));
    ss=*sp;
    ss->correct=correct;
    ss->size=size;
    ss->tree.label=label;
    ss->tree.left=left;
    ss->tree.right=right;
    ss->prob=prob;
    ss->next=cc->strat_list;  /* add to beginning of list */
    cc->strat_list=ss;      

    cc->tree.left=left;
    cc->tree.right=right;
    cc->prob=prob;

    return(1);
  }

  if (ss->prob < prob) { /* better than best in this strata so far */
    ss->tree.left=left;
    ss->tree.right=right;
    ss->prob=prob;
    if (cc->prob < prob) { /* better than best so far overall */
      cc->tree.left=left;
      cc->tree.right=right;
      cc->prob=prob;
    }
    return(1);
  }

  return(0);
}

int
add_edges(sihashcc chart_entry, si_index label, chart_cell cl, chart_cell cr,
	  double prob, int correct, int mode, int add_mode, si_t si)
{
  int changed=0;

  if (mode == CKY_MAX) {
    if(cr) {
      /* it is a binary rule */
      changed+=add_edge_zeroone(chart_entry, label, 
				&cl->tree, &cr->tree, 
				cl->prob + cr->prob + prob, add_mode, si);
      if (cr->present2)
	changed+=add_edge_zeroone(chart_entry, label, 
				  &cl->tree, &cr->tree2, 
				  cl->prob + cr->prob2 + prob, add_mode, si);
      if (cl->present2)
	changed+=add_edge_zeroone(chart_entry, label, 
				  &cl->tree2, &cr->tree, 
				  cl->prob2 + cr->prob + prob, add_mode, si);
    }
    else if(cl) {
      /* it is a unary rule */
	changed+=add_edge_zeroone(chart_entry, label, 
				  &cl->tree, NULL, 
				  cl->prob + prob, add_mode, si);
	if(cl->present2) 
	  changed+=add_edge_zeroone(chart_entry, label, 
				    &cl->tree2, NULL, 
				    cl->prob2 + prob, add_mode, si);
    }
    else {
      /* it is a terminal */
      changed+=add_edge_zeroone(chart_entry, label, NULL, NULL, 0.0, add_mode,
				si);
    }
  }
  else if (mode == CKY_STRATIFIED) {

    if(add_mode == NO_ADD_NEW_SYMBOLS) 
      if(!sihashcc_ref(chart_entry,label))
	return(0);

    strat_cell ls,rs;
    chart_cell *cp = sihashcc_valuep(chart_entry, label);
    chart_cell cc = *cp;
    if (cc == NULL) {  /* construct a new chart entry */
      *cp = make_chart_cell(label, NULL, NULL, 0.0, 1, 0, CKY_STRATIFIED);
      cc = *cp;
    }  
  
    if(cr) {
      /* it is a binary rule */
     for(ls=cl->strat_list;ls;ls=ls->next) {
	for(rs=cr->strat_list;rs;rs=rs->next) {
	  /* if((cc->prob-(ls->prob + rs->prob + prob)) < 1.0) * ONLY for mult-loss */
	    changed+=add_edge_stratified(cc, label, &ls->tree,
					 &rs->tree,
					 ls->prob + rs->prob + prob,
					 ls->correct + rs->correct + correct,
					 ls->size + rs->size + is_not_binarized(label, si), 
					 si);
	}
      }
    }
    else if(cl) {
      /* it is a unary rule */
      for(ls=cl->strat_list;ls;ls=ls->next) {
	/* if((cc->prob-(ls->prob + prob)) < 1.0)  ONLY for mult-loss */
	  changed+=add_edge_stratified(cc, label, &ls->tree, NULL,
				       ls->prob + prob,
				       ls->correct + correct, 
				       ls->size + is_not_binarized(label, si), 
				       si);
      }
    }
    else {
      /* it is a terminal */
      changed+=add_edge_stratified(cc, label, NULL, NULL, 0.0, 0, 0, si);
    }
  }

  return(changed);
}

int
apply_unary(sihashcc chart_entry, struct ledges *l, 
	    STRUCTMODEL *sm, si_t si, int mode, int add_mode, 
	    int lpos, int rpos, int start, int end)
{
  sihashursit	ursit;
  size_t	i;
  int           changed=0;
  int           correct=0;

  for (ursit=sihashursit_init(sm->grammar.urs); sihashursit_ok(ursit); 
       ursit = sihashursit_next(ursit)) {
    /* look up the rule's child category */
    chart_cell c = sihashcc_ref(chart_entry, ursit.key);	
    
    if (c) {			/* such categories exist in this cell */
      for (i=0; i<ursit.value.n; i++) {
	/* check if the label of rule is correct */
	if(l)
	  correct=member_ledges(l,remove_parent_from_label(ursit.value.e[i]->parent,si));

	changed+=add_edges(chart_entry, ursit.value.e[i]->parent, 
			   c, NULL, 
			   urule_value(ursit.value.e[i],&c->tree,sm,
				       lpos,rpos,start,end),
			   correct, mode, add_mode, si);
      }
    }
  }
  return(changed);
}

static void
apply_binary(sihashcc parent_entry, struct ledges *l,
	     sihashcc left_entry, sihashcc right_entry,
	     STRUCTMODEL *sm, si_t si, int mode, 
	     int lpos, int mpos, int rpos, int start, int end)
{
  sihashbrsit	brsit;
  size_t	i;
  int           correct=0;

  for (brsit=sihashbrsit_init(sm->grammar.brs); sihashbrsit_ok(brsit); 
       brsit = sihashbrsit_next(brsit)) {
    /* look up the rule's left category */
    chart_cell cl = sihashcc_ref(left_entry, brsit.key);
    if (cl)	/* such categories exist in this cell */
      for (i=0; i<brsit.value.n; i++) {
	chart_cell cr = sihashcc_ref(right_entry, brsit.value.e[i]->right);
	if (cr) {
	  /* check if the label of rule is correct */
	  if(l)
	    correct=member_ledges(l,remove_parent_from_label(brsit.value.e[i]->parent,si));
	  add_edges(parent_entry, brsit.value.e[i]->parent, 
		    cl, cr, 
		    brule_value(brsit.value.e[i],&cl->tree,&cr->tree,sm,
				lpos,mpos,rpos,start,end),
		    correct, mode, ADD_NEW_SYMBOLS, si);
	}
      }
  }
}

chart
cky(struct vindex terms, struct ledges *l, STRUCTMODEL *sm, si_t si, int mode)
{
  int left, right, mid, i;
  chart c;
  struct ledges *lsub=NULL;

  c = chart_make(terms.n);
  
  /* insert lexical items */
  for (left=0; left< (int) terms.n; left++) {
    right=left+1;
    si_index	label = terms.e[left];
    sihashcc	chart_entry = make_sihashcc(NLABELS);
    CHART_ENTRY(c, left, right) = chart_entry;
    add_edges(chart_entry, label, NULL, NULL, 0.0, 0, mode, ADD_NEW_SYMBOLS, 
	      si);
    /* 
       chart_cell  cell = make_chart_cell(label, NULL, NULL, 0.0, 1, 0, mode);
       sihashcc_set(chart_entry, label, cell);
    */

    /* find the correct labels for this ledge */
    if(l) lsub=filter_ledges(l,left,right);

    /* close under unary rules */
    for(i=0;(i<MAXUNARYDEPTH) && apply_unary(chart_entry,lsub,sm,si,mode,ADD_NEW_SYMBOLS,left,left+1,0,terms.n);i++);
    /* make sure there are no values that have not propagated all the way */
    for(i=0;apply_unary(chart_entry,lsub,sm,si,mode,NO_ADD_NEW_SYMBOLS,left,left+1,0,terms.n);i++);

    if(lsub) free_ledges(lsub);
  }

  for (right=2; right<=(int)terms.n; right++)
    for (left=right-2; left>=0; left--) {
      sihashcc chart_entry = make_sihashcc(CHART_CELLS);   
      CHART_ENTRY(c, left, right) = chart_entry;

      /* find the correct labels for this ledge */
      if(l) lsub=filter_ledges(l,left,right);

      for (mid=left+1; mid<right; mid++) 
	apply_binary(chart_entry, lsub, CHART_ENTRY(c,left,mid), 
		     CHART_ENTRY(c,mid,right), sm, si, mode,
		     left,mid,right,0,terms.n);

      for(i=0;(i<MAXUNARYDEPTH) && apply_unary(chart_entry,lsub,sm,si,mode,ADD_NEW_SYMBOLS,left,right,0,terms.n);i++);
      for(i=0;apply_unary(chart_entry,lsub,sm,si,mode,NO_ADD_NEW_SYMBOLS,left,right,0,terms.n);i++);

      if(lsub) free_ledges(lsub);
    }

  return c;
}

chart cky_wrap(struct vindex terms, tree correct_tree, STRUCTMODEL *sm, 
	       si_t si)
{
  int i=0,first=1;
  chart_cell root_cell;
  double bestprob=0;
  struct ledges *l;
  tree t;
  strat_cell ls;

  if(correct_tree) {
    t = remove_parent_annotation(correct_tree, si); /* just in case */
    l = tree_ledges(t);                             /* get correct ledges */
    free_tree(t);
  }
  else {
    l = NULL;
  }
  chart c=cky(terms, l, sm, si, CKY_STRATIFIED);

  root_cell = sihashcc_ref(CHART_ENTRY(c, 0, terms.n),
			   si_string_index(si, ROOT));
  if(root_cell) {
    for(ls=root_cell->strat_list;ls;ls=ls->next) {
      if((ls->prob > bestprob) || first) {
	bestprob=ls->prob;
	root_cell->tree=ls->tree;
	root_cell->prob=ls->prob;
	first=0;
      }
      i++;
    }
    /* printf("strat_n=%d ",i); */
  }

  if(l) free_ledges(l);

  return(c);
} 

chart cky_maxloss(struct vindex terms, PATTERN x, LABEL y,  
		  STRUCTMODEL *sm, STRUCT_LEARN_PARM *sparm, 
		  double *maxloss)
{
  int first=1;
  int maxcorrect=0,maxsize=0;
  chart_cell root_cell;
  double maxviol;
  struct ledges *l;
  tree t;
  FILE    *tracefp=NULL;
  strat_cell ls;

  SVECTOR *v=psi(x,y,sm,sparm);
  DOC *ex=create_example(-1,-1,-1,1,v);
  double yprob=classify_example(sm->svm_model,ex);
  free_example(ex,1);

  t = remove_parent_annotation(y.parse, sm->si); /* just in case */
  l = tree_ledges(t);                            /* get correct ledges */
  free_tree(t);

  chart c=cky(terms, l, sm, sm->si, CKY_STRATIFIED);

  root_cell = sihashcc_ref(CHART_ENTRY(c, 0, terms.n),
			   si_string_index(sm->si, ROOT));
  (*maxloss)=0.0;
  maxviol=0.0;
  if(root_cell) {
    for(ls=root_cell->strat_list;ls;ls=ls->next) {
      double prob=ls->prob;
      int correct=ls->correct;
      int size=ls->size;
      double loss=0;
      double viol=0;
      double chi=(yprob-prob);
      if((sparm->loss_function == 1) 
	 && (sparm->loss_type == SLACK_RESCALING) 
	 && (sparm->slack_norm == 1)) {
	loss=(1.0-fone(correct,l->n,size));
	viol=(1.0-chi)*loss;
      }
      else if((sparm->loss_function == 1) 
	      && (sparm->loss_type == SLACK_RESCALING) 
	      && (sparm->slack_norm == 2)) {
	loss=sqrt(1.0-fone(correct,l->n,size));
	viol=(1.0-chi)*loss;
      }
      else if((sparm->loss_function == 1) 
	      && (sparm->loss_type == MARGIN_RESCALING) 
	      && (sparm->slack_norm == 1)) {
	loss=(1.0-fone(correct,l->n,size));
	viol=loss-chi;
      }
      else if((sparm->loss_function == 1) 
	      && (sparm->loss_type == MARGIN_RESCALING) 
	      && (sparm->slack_norm == 2)) {
	loss=sqrt(1.0-fone(correct,l->n,size));
	viol=loss-chi;
      }
      else {
	assert(0);
      }
      if(viol<0.0) viol=0.0;
      if((viol > maxviol) || first) {
	maxviol=viol;
	(*maxloss)=loss;
	maxcorrect=correct;
	maxsize=size;
	root_cell->tree=ls->tree;
	root_cell->prob=prob;
	first=0;
      }
    }
  }

  if(tracefp)
    fprintf(tracefp,"return: maxviol %g with loss %g at %d/%d of %d\n",maxviol,(*maxloss),maxcorrect,maxsize,l->n); 
  
  free_ledges(l);

  return(c);
} 

double fone(int correct, int trueledgenum, int predledgenum) 
{
  if((trueledgenum == 0) || (predledgenum == 0))
    return(0.0);
  double precision=(double)correct/predledgenum;
  double recall=(double)correct/trueledgenum;
  return(2.0*precision*recall/(precision+recall));
}
